# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import
from __future__ import print_function
import numpy as np
import warnings

from .material import Material, material_property
from .mineral import Mineral
from . import averaging_schemes
from . import chemicalpotentials
from .solidsolution import SolidSolution
from .processchemistry import sum_formulae

def check_pairs(phases, fractions):
        if len(fractions) < 1:
            raise Exception('ERROR: we need at least one phase')

        if len(phases) != len(fractions):
            raise Exception(
                'ERROR: different array lengths for phases and fractions')

        total = sum(fractions)
        if abs(total - 1.0) > 1e-10:
            raise Exception(
                'ERROR: list of molar fractions does not add up to one')
        for p in phases:
            if not isinstance(p, Mineral):
                raise Exception(
                    'ERROR: object of type ''%s'' is not of type Mineral' % (type(p)))


# static composite of minerals/composites
class Composite(Material):

    """
    Base class for a composite material.
    The static phases can be minerals or materials,
    meaning composite can be nested arbitrarily.

    The fractions of the phases can be input
    as either 'molar' or 'mass' during instantiation,
    and modified (or initialised) after this point by
    using set_fractions.

    This class is available as ``burnman.Composite``.
    """

    def __init__(self, phases, fractions=None, fraction_type='molar', name='Unnamed composite'):
        """
        Create a composite using a list of phases and their fractions (adding to 1.0).

        Parameters
        ----------
        phases: list of :class:`burnman.Material`
            list of phases.
        fractions: list of floats
            molar or mass fraction for each phase.
        fraction_type: 'molar' or 'mass' (optional, 'molar' as standard)
            specify whether molar or mass fractions are specified.
        """

        Material.__init__(self)

        assert(len(phases) > 0)
        self.phases = phases

        if fractions is not None:
            self.set_fractions(fractions, fraction_type)
        else:
            self.molar_fractions = None

        self.set_averaging_scheme('VoigtReussHill')
        self.name=name
        self.print_precision=4 # number of significant figures used by self.__str__

    def __str__(self):
        string='Composite: {0}'.format(self.name)
        try:
            string += '\n  P, T: {0:.{sf}g} Pa, {1:.{sf}g} K'.format(self.pressure,
                                                                     self.temperature,
                                                                     sf=self.print_precision)
        except:
            pass
        string+='\nPhase and endmember fractions:'
        for phase, fraction in zip(*self.unroll()):
            string+='\n  {0}: {1:0.{sf}f}'.format(phase.name, fraction, sf=self.print_precision)
            if isinstance(phase, SolidSolution):
               for i in range(phase.n_endmembers):
                   string+='\n    {0}: {1:0.{sf}f}'.format(phase.endmember_names[i],
                                                           phase.molar_fractions[i],
                                                           sf=self.print_precision)
        return string

    def set_fractions(self, fractions, fraction_type='molar'):
        """
        Change the fractions of the phases of this Composite.
        Resets cached properties

        Parameters
        ----------
        fractions: list of floats
            molar or mass fraction for each phase.
        fraction_type: 'molar' or 'mass'
            specify whether molar or mass fractions are specified.
        """
        assert(len(self.phases) == len(fractions))

        try:
            total = sum(fractions)
        except TypeError:
            raise Exception(
                "Since v0.8, burnman.Composite takes an array of Materials, then an array of fractions")

        for f in fractions:
            assert (f >= -1e-12)

        self.reset()

        if abs(total - 1.0) > 1e-12:
            warnings.warn(
                "Warning: list of fractions does not add up to one but %g. Normalizing." % total)
            corrected_fractions = [fr / total for fr in fractions]
            fractions = corrected_fractions

        if fraction_type == 'molar':
            molar_fractions = fractions
        elif fraction_type == 'mass':
            molar_fractions = self._mass_to_molar_fractions(
                self.phases, fractions)
        else:
            raise Exception(
                "Fraction type not recognised. Please use 'molar' or mass")

        # Set minimum value of a molar fraction at 0.0 (rather than -1.e-12)
        self.molar_fractions = [max(0.0, fraction)
                                for fraction in molar_fractions]

    def set_method(self, method):
        """
        set the same equation of state method for all the phases in the composite
        """
        for phase in self.phases:
            phase.set_method(method)
        # Clear the cache on resetting method
        self.reset()

    def set_averaging_scheme(self, averaging_scheme):
        """
        Set the averaging scheme for the moduli in the composite.
        Default is set to VoigtReussHill, when Composite is initialized.
        """

        if type(averaging_scheme) == str:
            self.averaging_scheme = getattr(
                averaging_schemes, averaging_scheme)()
        else:
            self.averaging_scheme = averaging_scheme
        # Clear the cache on resetting averaging scheme
        self.reset()

    def set_state(self, pressure, temperature):
        """
        Update the material to the given pressure [Pa] and temperature [K].
        """
        Material.set_state(self, pressure, temperature)
        for phase in self.phases:
            phase.set_state(pressure, temperature)

    def debug_print(self, indent=""):
        print("{0}Composite: {1}".format(indent, self.name))
        indent += "  "
        if self.molar_fractions is None:
            for i, phase in enumerate(self.phases):
                phase.debug_print(indent + "  ")
        else:
            for i, phase in enumerate(self.phases):
                print("%s%g of" % (indent, self.molar_fractions[i]))
                phase.debug_print(indent + "  ")

    def unroll(self):
        if self.molar_fractions is None:
            raise Exception(
                "Unroll only works if the composite has defined fractions.")
        phases = []
        fractions = []
        for i, phase in enumerate(self.phases):
            p_mineral, p_fraction = phase.unroll()
            check_pairs(p_mineral, p_fraction)
            fractions.extend([f * self.molar_fractions[i] for f in p_fraction])
            phases.extend(p_mineral)
        return phases, fractions

    def to_string(self):
        """
        return the name of the composite
        """
        return "{0}: {1}".format(self.__class__.__name__, self.name)

    @material_property
    def formula(self):
        """
        Returns molar chemical formula of the composite
        """
        return sum_formulae([ph.formula for ph in self.phases], self.molar_fractions)

    @material_property
    def molar_internal_energy(self):
        """
        Returns molar internal energy of the mineral [J/mol]
        Aliased with self.energy
        """
        U = sum(phase.molar_internal_energy * molar_fraction for (
                phase, molar_fraction) in zip(self.phases, self.molar_fractions))
        return U

    @material_property
    def molar_gibbs(self):
        """
        Returns molar Gibbs free energy of the composite [J/mol]
        Aliased with self.gibbs
        """
        G = sum(phase.molar_gibbs * molar_fraction for (phase, molar_fraction)
                in zip(self.phases, self.molar_fractions))
        return G

    @material_property
    def molar_helmholtz(self):
        """
        Returns molar Helmholtz free energy of the mineral [J/mol]
        Aliased with self.helmholtz
        """
        F = sum(phase.molar_helmholtz * molar_fraction for (
                phase, molar_fraction) in zip(self.phases, self.molar_fractions))
        return F

    @material_property
    def molar_volume(self):
        """
        Returns molar volume of the composite [m^3/mol]
        Aliased with self.V
        """
        volumes = np.array(
            [phase.molar_volume * molar_fraction for (phase, molar_fraction) in zip(self.phases, self.molar_fractions)])
        return np.sum(volumes)

    @material_property
    def molar_mass(self):
        """
        Returns molar mass of the composite [kg/mol]
        """
        return sum([phase.molar_mass * molar_fraction for (phase, molar_fraction) in zip(self.phases, self.molar_fractions)])

    @material_property
    def density(self):
        """
        Compute the density of the composite based on the molar volumes and masses
        Aliased with self.rho
        """
        densities = np.array([phase.density for phase in self.phases])
        volumes = np.array(
            [phase.molar_volume * molar_fraction for (phase, molar_fraction) in zip(self.phases, self.molar_fractions)])
        return self.averaging_scheme.average_density(volumes, densities)

    @material_property
    def molar_entropy(self):
        """
        Returns enthalpy of the mineral [J]
        Aliased with self.S
        """
        S = sum(phase.molar_entropy * molar_fraction for (
                phase, molar_fraction) in zip(self.phases, self.molar_fractions))
        return S

    @material_property
    def molar_enthalpy(self):
        """
        Returns enthalpy of the mineral [J]
        Aliased with self.H
        """
        H = sum(phase.molar_enthalpy * molar_fraction for (
                phase, molar_fraction) in zip(self.phases, self.molar_fractions))
        return H

    @material_property
    def isothermal_bulk_modulus(self):
        """
        Returns isothermal bulk modulus of the composite [Pa]
        Aliased with self.K_T
        """
        V_frac = np.array([phase.molar_volume * molar_fraction for (
                           phase, molar_fraction) in zip(self.phases, self.molar_fractions)])
        K_ph = np.array(
            [phase.isothermal_bulk_modulus for phase in self.phases])
        G_ph = np.array([phase.shear_modulus for phase in self.phases])

        return self.averaging_scheme.average_bulk_moduli(V_frac, K_ph, G_ph)

    @material_property
    def adiabatic_bulk_modulus(self):
        """
        Returns adiabatic bulk modulus of the mineral [Pa]
        Aliased with self.K_S
        """
        V_frac = np.array([phase.molar_volume * molar_fraction for (
                           phase, molar_fraction) in zip(self.phases, self.molar_fractions)])
        K_ph = np.array(
            [phase.adiabatic_bulk_modulus for phase in self.phases])
        G_ph = np.array([phase.shear_modulus for phase in self.phases])

        return self.averaging_scheme.average_bulk_moduli(V_frac, K_ph, G_ph)

    @material_property
    def isothermal_compressibility(self):
        """
        Returns isothermal compressibility of the composite (or inverse isothermal bulk modulus) [1/Pa]
        Aliased with self.beta_T
        """
        return 1. / self.isothermal_bulk_modulus

    @material_property
    def adiabatic_compressibility(self):
        """
        Returns isothermal compressibility of the composite (or inverse isothermal bulk modulus) [1/Pa]
        Aliased with self.beta_S
        """
        return 1. / self.adiabatic_bulk_modulus

    @material_property
    def shear_modulus(self):
        """
        Returns shear modulus of the mineral [Pa]
        Aliased with self.G
        """
        V_frac = np.array([phase.molar_volume * molar_fraction for (
                           phase, molar_fraction) in zip(self.phases, self.molar_fractions)])
        K_ph = np.array(
            [phase.adiabatic_bulk_modulus for phase in self.phases])
        G_ph = np.array([phase.shear_modulus for phase in self.phases])

        return self.averaging_scheme.average_shear_moduli(V_frac, K_ph, G_ph)

    @material_property
    def p_wave_velocity(self):
        """
        Returns P wave speed of the composite [m/s]
        Aliased with self.v_p
        """
        return np.sqrt((self.adiabatic_bulk_modulus + 4. / 3. *
                        self.shear_modulus) / self.density)

    @material_property
    def bulk_sound_velocity(self):
        """
        Returns bulk sound speed of the composite [m/s]
        Aliased with self.v_phi
        """
        return np.sqrt(self.adiabatic_bulk_modulus / self.density)

    @material_property
    def shear_wave_velocity(self):
        """
        Returns shear wave speed of the composite [m/s]
        Aliased with self.v_s
        """
        return np.sqrt(self.shear_modulus / self.density)

    @material_property
    def grueneisen_parameter(self):
        """
        Returns grueneisen parameter of the composite [unitless]
        Aliased with self.gr
        """
        return self.thermal_expansivity * self.isothermal_bulk_modulus * self.molar_volume / self.molar_heat_capacity_v

    @material_property
    def thermal_expansivity(self):
        """
        Returns thermal expansion coefficient of the composite [1/K]
        Aliased with self.alpha
        """
        volumes = np.array(
            [phase.molar_volume * molar_fraction for (phase, molar_fraction) in zip(self.phases, self.molar_fractions)])
        alphas = np.array([phase.thermal_expansivity for phase in self.phases])
        return self.averaging_scheme.average_thermal_expansivity(volumes, alphas)

    @material_property
    def molar_heat_capacity_v(self):
        """
        Returns molar_heat capacity at constant volume of the composite [J/K/mol]
        Aliased with self.C_v
        """
        c_v = np.array([phase.molar_heat_capacity_v for phase in self.phases])
        return self.averaging_scheme.average_heat_capacity_v(self.molar_fractions, c_v)

    @material_property
    def molar_heat_capacity_p(self):
        """
        Returns molar heat capacity at constant pressure of the composite [J/K/mol]
        Aliased with self.C_p
        """
        c_p = np.array([phase.molar_heat_capacity_p for phase in self.phases])
        return self.averaging_scheme.average_heat_capacity_p(self.molar_fractions, c_p)

    def _mass_to_molar_fractions(self, phases, mass_fractions):
        """
        Converts a set of mass fractions for phases into a set of molar fractions.

        Parameters
        ----------
        phases : list of :class:`burnman.Material`
        The list of phases for which fractions should be converted.

        mass_fractions : list of floats
        The list of mass fractions of the input phases.

        Returns
        -------
        molar_fractions : list of floats
        The list of molar fractions corresponding to the input molar fractions
        """
        total_moles = sum(
            mass_fraction / phase.molar_mass for mass_fraction, phase in zip(mass_fractions, phases))
        molar_fractions = [mass_fraction / (phase.molar_mass * total_moles)
                           for mass_fraction, phase in zip(mass_fractions, phases)]
        return molar_fractions



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

# This module provides the functions required to process the
# standard burnman formula formats.
# ProcessChemistry returns the number of atoms and molar mass of a compound
# given its unit formula as an argument.
# process_solution_chemistry returns information required to calculate
# solid solution properties from a set of endmember formulae

from __future__ import absolute_import
import re
import numpy as np
from fractions import Fraction
from collections import Counter
import pkgutil
from string import ascii_uppercase as ucase
from sympy import nsimplify


def read_masses():
    """
    A simple function to read a file with a two column list of
    elements and their masses into a dictionary
    """
    datastream = pkgutil.get_data(
        'burnman', 'data/input_masses/atomic_masses.dat')
    datalines = [line.strip()
                 for line in datastream.decode('ascii').split('\n')
                 if line.strip()]
    lookup = dict()
    for line in datalines:
        data = "%".join(line.split("%")[:1]).split()
        if data != []:
            lookup[data[0]] = float(data[1])
    return lookup


"""
atomic_masses is a dictionary of atomic masses
"""
atomic_masses = read_masses()

"""
IUPAC_element_order provides a list of all the elements.
Element order is based loosely on electronegativity,
following the scheme suggested by IUPAC, except that H
comes after the Group 16 elements, not before them.
"""
IUPAC_element_order = ['v', 'Og', 'Rn', 'Xe', 'Kr', 'Ar', 'Ne', 'He',  # Group 18
                       'Fr', 'Cs', 'Rb', 'K', 'Na', 'Li',  # Group 1 (not H)
                       'Ra', 'Ba', 'Sr', 'Ca', 'Mg', 'Be',  # Group 2
                       'Lr', 'No', 'Md', 'Fm', 'Es', 'Cf', 'Bk', 'Cm',
                       'Am', 'Pu', 'Np', 'U', 'Pa', 'Th', 'Ac',  # Actinides
                       'Lu', 'Yb', 'Tm', 'Er', 'Ho', 'Dy', 'Tb', 'Gd',
                       'Eu', 'Sm', 'Pm', 'Nd', 'Pr', 'Ce', 'La',  # Lanthanides
                       'Y', 'Sc',  # Group 3
                       'Rf', 'Hf', 'Zr', 'Ti',  # Group 4
                       'Db', 'Ta', 'Nb', 'V',  # Group 5
                       'Sg', 'W', 'Mo', 'Cr',  # Group 6
                       'Bh', 'Re', 'Tc', 'Mn',  # Group 7
                       'Hs', 'Os', 'Ru', 'Fe',  # Group 8
                       'Mt', 'Ir', 'Rh', 'Co',  # Group 9
                       'Ds', 'Pt', 'Pd', 'Ni',  # Group 10
                       'Rg', 'Au', 'Ag', 'Cu',  # Group 11
                       'Cn', 'Hg', 'Cd', 'Zn',  # Group 12
                       'Nh', 'Tl', 'In', 'Ga', 'Al', 'B',  # Group 13
                       'Fl', 'Pb', 'Sn', 'Ge', 'Si', 'C',  # Group 14
                       'Mc', 'Bi', 'Sb', 'As', 'P', 'N',  # Group 15
                       'Lv', 'Po', 'Te', 'Se', 'S', 'O',  # Group 16
                       'H',  # hydrogen
                       'Ts', 'At', 'I', 'Br', 'Cl', 'F']  # Group 17


def dictionarize_formula(formula):
    """
    A function to read a chemical formula string and
    convert it into a dictionary

    Parameters
    ----------
    formula : string object
        Chemical formula, written in the XnYm format, where
        the formula has n atoms of element X and m atoms of element Y

    Returns
    -------
    f : dictionary object
        The same chemical formula, but expressed as a dictionary.
    """
    f = dict()
    elements = re.findall('[A-Z][^A-Z]*', formula)
    for element in elements:
        element_name = re.split('[0-9][^A-Z]*', element)[0]
        element_atoms = re.findall('[0-9][^A-Z]*', element)
        if len(element_atoms) == 0:
            element_atoms = Fraction(1.0)
        else:
            element_atoms = Fraction(element_atoms[0])
        f[element_name] = f.get(element_name, 0.0) + element_atoms

    return f


def sum_formulae(formulae, amounts=None):
    """
    Adds together a set of formulae.

    Parameters
    ----------
    formulae : list of dictionary or counter objects
        List of chemical formulae
    amounts : list of floats
        List of amounts of each formula

    Returns
    -------
    summed_formula : Counter object
        The sum of the user-provided formulae
    """
    if amounts is None:
        amounts = [1. for formula in formulae]
    else:
        assert (len(formulae) == len(amounts))

    summed_formula = Counter()
    for i, formula in enumerate(formulae):
        summed_formula.update(Counter({element: amounts[i] * n_atoms
                                       for (element, n_atoms)
                                       in formula.items()}))
    return summed_formula


def formula_mass(formula):
    """
    A function to take a chemical formula and compute the formula mass.

    Parameters
    ----------
    formula : dictionary or counter object
        A chemical formula

    Returns
    -------
    mass : float
        The mass per mole of formula
    """
    mass = sum(formula[element] * atomic_masses[element]
               for element in formula)
    return mass


def convert_formula(formula, to_type='mass', normalize=False):
    """
    Converts a chemical formula from one type (mass or molar)
    into the other. Renormalises amounts if normalize=True

    Parameters
    ----------
    formula : dictionary or counter object
        A chemical formula

    to_type : string, one of 'mass' or 'molar'
        Conversion type

    normalize : boolean
        Whether or not to normalize the converted formula to 1

    Returns
    -------
    f : dictionary
        The converted formula
    """

    if to_type == 'mass':
        f = {element: n_atoms * atomic_masses[element]
             for (element, n_atoms) in formula.items()}
    elif to_type == 'molar':
        f = {element: n_atoms / atomic_masses[element]
             for (element, n_atoms) in formula.items()}
    else:
        raise Exception('Value of parameter to_type not recognised. '
                        'Should be either "mass" or "molar".')

    if normalize:
        s = np.sum([n for (element, n) in f.items()])
        f = {element: n/s for (element, n) in f.items()}

    return f


def process_solution_chemistry(solution_model):
    """
    This function parses a class instance with a "formulas"
    attribute containing site information, e.g.

        [ '[Mg]3[Al]2Si3O12', '[Mg]3[Mg1/2Si1/2]2Si3O12' ]

    It outputs the bulk composition of each endmember
    (removing the site information), and also a set of
    variables and arrays which contain the site information.
    These are output in a format that can easily be used to
    calculate activities and gibbs free energies, given
    molar fractions of the phases and pressure
    and temperature where necessary.

    Parameters
    ----------
    solution_model : instance of class
        Class must have a "formulas" attribute, containing a
        list of chemical formulae with site information

    Creates attributes
    ------------------
    solution_formulae : list of dictionaries
        List of endmember formulae is output from site formula strings

    n_sites : integer
        Number of sites in the solid solution.
        Should be the same for all endmembers.

    sites : list of lists of strings
        A list of elements for each site in the solid solution

    site_names : list of strings
        A list of elements_site pairs in the solid solution, where
        each distinct site is given by a unique uppercase letter
        e.g. ['Mg_A', 'Fe_A', 'Al_A', 'Al_B', 'Si_B']

    n_occupancies : integer
        Sum of the number of possible elements on each of the sites
        in the solid solution.
        Example: A binary solution [[A][B],[B][C1/2D1/2]] would have
        n_occupancies = 5, with two possible elements on
        Site 1 and three on Site 2

    site_multiplicities : array of floats
        The number of each site per formula unit
        To simplify computations later, the multiplicities
        are repeated for each element on each site

    endmember_occupancies : 2d array of floats
        A 1D array for each endmember in the solid solution,
        containing the fraction of atoms of each element on each site.

    endmember_noccupancies : 2d array of floats
        A 1D array for each endmember in the solid solution,
        containing the number of atoms of each element on each site
        per mole of endmember.
    """
    formulae = solution_model.formulas
    n_sites = formulae[0].count('[')
    n_endmembers = len(formulae)

    # Check the number of sites is the same for all endmembers
    for i in range(n_endmembers):
        assert(formulae[i].count('[') == n_sites)

    solution_formulae = []
    sites = [[] for i in range(n_sites)]
    list_occupancies = []
    list_multiplicity = np.empty(shape=(n_sites))
    n_occupancies = 0

    # Number of unique site occupancies (e.g.. Mg on X etc.)
    for i_mbr in range(n_endmembers):
        solution_formula = dict()
        list_occupancies.append([[0] * len(sites[site])
                                for site in range(n_sites)])
        s = re.split(r'\[', formulae[i_mbr])[1:]

        for i_site, site_string in enumerate(s):
            site_split = re.split(r'\]', site_string)
            site_occupancy = site_split[0]

            mult = re.split('[A-Z][^A-Z]*', site_split[1])[0]
            if mult == '':
                list_multiplicity[i_site] = Fraction(1.0)
            else:
                list_multiplicity[i_site] = Fraction(mult)

            # Loop over elements on a site
            elements = re.findall('[A-Z][^A-Z]*', site_occupancy)

            for element in elements:
                # Find the element and proportion on the site
                element_split = re.split('([0-9][^A-Z]*)', element)
                element_on_site = element_split[0]
                if len(element_split) == 1:
                    proportion_element_on_site = Fraction(1.0)
                else:
                    proportion_element_on_site = Fraction(element_split[1])

                solution_formula[element_on_site] = solution_formula.get(
                    element_on_site, 0.0) + (list_multiplicity[i_site]
                                             * proportion_element_on_site)

                if element_on_site not in sites[i_site]:
                    n_occupancies += 1
                    sites[i_site].append(element_on_site)
                    i_el = sites[i_site].index(element_on_site)
                    for parsed_mbr in range(len(list_occupancies)):
                        list_occupancies[parsed_mbr][i_site].append(0)
                else:
                    i_el = sites[i_site].index(element_on_site)
                list_occupancies[i_mbr][i_site][i_el] = proportion_element_on_site

            # Loop over elements after site
            if len(site_split) != 1:
                not_in_site = str(filter(None, site_split[1]))
                not_in_site = not_in_site.replace(mult, '', 1)
                for enamenumber in re.findall('[A-Z][^A-Z]*', not_in_site):
                    element = list(
                        filter(None, re.split(r'(\d+)', enamenumber)))
                    # Look up number of atoms of element
                    if len(element) == 1:
                        nel = 1.
                    else:
                        nel = float(float(element[1]))
                    solution_formula[element[0]] = solution_formula.get(
                        element[0], 0.0) + nel

        solution_formulae.append(solution_formula)

    # Site occupancies and multiplicities
    endmember_occupancies = np.empty(shape=(n_endmembers, n_occupancies))
    site_multiplicities = np.empty(shape=(n_occupancies))
    for i_mbr in range(n_endmembers):
        n_element = 0
        for i_site in range(n_sites):
            for i_el in range(len(list_occupancies[i_mbr][i_site])):
                endmember_occupancies[i_mbr][
                    n_element] = list_occupancies[i_mbr][i_site][i_el]
                site_multiplicities[n_element] = list_multiplicity[i_site]
                n_element += 1

    # Site names
    solution_model.site_names = []
    for i, elements in enumerate(sites):
        for element in elements:
            solution_model.site_names.append('{0}_{1}'.format(element,
                                                              ucase[i]))

    # Finally, make attributes for solution model instance:
    solution_model.solution_formulae = solution_formulae
    solution_model.n_sites = n_sites
    solution_model.sites = sites
    solution_model.n_occupancies = n_occupancies
    solution_model.site_multiplicities = site_multiplicities
    solution_model.endmember_occupancies = endmember_occupancies
    solution_model.endmember_noccupancies = np.einsum('ij, j->ij',
                                                      endmember_occupancies,
                                                      site_multiplicities)


def site_occupancies_to_strings(site_species_names, site_multiplicities,
                                endmember_occupancies):
    """
    Converts a list of endmember site occupancies into a list
    of string representations of those occupancies.

    Parameters
    ----------
    site_species_names : 2D list of strings
        A list of list of strings, giving the names of the species
        which reside on each site.
        List of sites, each of which contains a list of the species
        occupying each site.

    site_multiplicities : numpy array of floats
        List of floats giving the multiplicity of each site
        Must be either the same length as the number of sites, or
        the same length as site_species_names
        (with an implied repetition of the same
        number for each species on a given site).

    endmember_occupancies : 2D numpy array of floats
        A list of site-species occupancies for each endmember.
        The first dimension loops over the endmembers, and the
        second dimension loops over the site-species occupancies for that endmember.
        The total number and order of occupancies must
        be the same as the strings in site_species_names.

    Returns
    -------
    site_formulae : list of strings
        A list of strings in standard burnman format.
        For example, [Mg]3[Al]2 would correspond to the
        classic two-site pyrope garnet.
    """

    # Site multiplicities should either be given on a per-site basis,
    # or a per-species basis
    if len(site_species_names) == len(site_multiplicities):
        site_mults = []

        for i, site in enumerate(site_species_names):
            for species in site:
                site_mults.append(site_multiplicities[i])

        site_multiplicities = site_mults

    elif len(endmember_occupancies[0]) != len(site_multiplicities):
        raise Exception(
            'Site multiplicities should either be given on a per-site basis or a per-species basis')

    site_formulae = []
    for mbr_occupancies in endmember_occupancies:
        i = 0
        site_formulae.append('')
        for site in site_species_names:
            amounts = mbr_occupancies[i:i+len(site)]
            mult = site_multiplicities[i]
            if np.abs(mult - 1.) < 1.e-12:
                mult = ''
            else:
                mult = str(nsimplify(mult))
            amounts /= sum(amounts)
            site_occupancy = formula_to_string(dict(zip(site, amounts)))
            site_formulae[-1] += '[{0}]{1}'.format(site_occupancy, mult)
            i += len(site)

    return site_formulae


def compositional_array(formulae):
    """
    Parameters
    ----------
    formulae : list of dictionaries
        List of chemical formulae

    Returns
    -------
    formula_array : 2D array of floats
        Array of endmember formulae

    elements : List of strings
        List of elements
    """
    elements = []
    for formula in formulae:
        for element in formula:
            if element not in elements:
                elements.append(element)

    formula_array = ordered_compositional_array(formulae, elements)

    return formula_array, elements


def ordered_compositional_array(formulae, elements):
    """
    Parameters
    ----------
    formulae : list of dictionaries
        List of chemical formulae

    elements : List of strings
        List of elements

    Returns
    -------
    formula_array : 2D array of floats
        Array of endmember formulae
    """
    formula_array = np.zeros(shape=(len(formulae), len(elements)))
    for idx, formula in enumerate(formulae):
        for element in formula:
            assert(element in elements)
            formula_array[idx][elements.index(element)] = formula[element]

    return formula_array


def formula_to_string(formula):
    """
    Parameters
    ----------
    formula : dictionary or counter
        Chemical formula

    Returns
    -------
    formula_string : string
        A formula string, with element order as given in the list
        IUPAC_element_order.
        If one or more keys in the dictionary are not one of the elements
        in the periodic table, then they are added at the end of the string.
    """

    formula_string = ''
    for e in IUPAC_element_order:
        if e in formula and np.abs(formula[e]) > 1.e-12:
            if np.abs(formula[e] - 1.) < 1.e-12:
                formula_string += e
            else:
                formula_string += e + str(nsimplify(formula[e]))

    for e in formula:
        if e not in IUPAC_element_order:
            if e in formula and np.abs(formula[e]) > 1.e-12:
                if np.abs(formula[e] - 1.) < 1.e-12:
                    formula_string += e
                else:
                    formula_string += e + str(nsimplify(formula[e]))

    return formula_string



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import
from __future__ import print_function

import operator
import bisect
import pkgutil
import numpy as np
from scipy.optimize import fsolve
from scipy.ndimage.filters import gaussian_filter
from scipy.interpolate import interp2d
import scipy.integrate as integrate
from collections import Counter
import itertools
import warnings

from . import constants


def copy_documentation(copy_from):
    """
    Decorator @copy_documentation(another_function) will copy the documentation found in a different
    function (for example from a base class). The docstring applied to some function a() will be ::

        (copied from BaseClass.some_function):
        <documentation from BaseClass.some_function>
        <optionally the documentation found in a()>

    """
    def mydecorator(func):
        def wrapper(*args):
            return func(*args)
        old = ""
        if func.__doc__:
            old = "\n" + func.__doc__

        copied_from = ""
        if hasattr(copy_from, "__name__"):
            copied_from = "(copied from " + copy_from.__name__ + "):\n"
        wrapper.__doc__ = copied_from + copy_from.__doc__ + old
        wrapper.__name__ = func.__name__
        return wrapper
    return mydecorator


def flatten(arr):
    return (flatten(arr[0]) + (flatten(arr[1:]) if len(arr) > 1 else [])
            if type(arr) is list or type(arr) is np.ndarray else [arr])


def round_to_n(x, xerr, n):
    return round(x, -int(np.floor(np.log10(np.abs(xerr)))) + (n - 1))


def unit_normalize(a, order=2, axis=-1):
    """
    Calculates the L2 normalized array of numpy array a
    of a given order and along a given axis.
    """
    l2 = np.atleast_1d(np.apply_along_axis(np.linalg.norm, axis, a, order))

    l2[l2 == 0] = 1
    return a / np.expand_dims(l2, axis)[0][0]


def pretty_print_values(popt, pcov, params):
    """
    Takes a numpy array of parameters, the corresponding covariance matrix
    and a set of parameter names and prints the parameters and
    principal 1-s.d.uncertainties (np.sqrt(pcov[i][i]))
    in a nice text based format.
    """
    for i, p in enumerate(params):
        p_rnd = round_to_n(popt[i], np.sqrt(pcov[i][i]), 1)
        c_rnd = round_to_n(np.sqrt(pcov[i][i]), np.sqrt(pcov[i][i]), 1)

        if p_rnd != 0.:
            p_expnt = np.floor(np.log10(np.abs(p_rnd)))
        else:
            p_expnt = 0.

        scale = np.power(10., p_expnt)
        nd = p_expnt - np.floor(np.log10(np.abs(c_rnd)))
        print('{0:s}: ({1:{4}{5}f} +/- {2:{4}{5}f}) x {3:.0e}'.format(p, p_rnd/scale, c_rnd/scale, scale, 0, (nd)/10.))


def pretty_print_table(table, use_tabs=False):
    """
    Takes a 2d table and prints it in a nice text based format. If
    use_tabs=True then only \t is used as a separator. This is useful for
    importing the data into other apps (Excel, ...). The default is to pad
    the columns with spaces to make them look neat. The first column is
    left aligned, while the remainder is right aligned.
    """
    if use_tabs:
        for r in table:
            print("\t".join(r).replace("_", "\_"))
        return

    def col_width(table, colidx):
        return max([len(str(row[colidx])) for row in table])

    # create a format string with the first column left aligned, the others right
    # example:   {:<27}{:>11}{:>6}{:>8}
    frmt = "".join(
        [('{:<' if i == 0 else '{:>') + str(1 + col_width(table, i)) + '}' for i in range(len(table[0]))])
    for r in table:
        print(frmt.format(*r))


def pretty_plot():
    """
    Makes pretty plots. Overwrites the matplotlib default settings to allow
    for better fonts. Slows down plotting
    """
    import matplotlib.pyplot as plt
    plt.rc('text', usetex=True)
    plt.rcParams['text.latex.preamble'] = '\\usepackage{relsize}'
    plt.rc('font', family='sanserif')


def sort_table(table, col=0):
    """
    Sort the table according to the column number
    """
    return sorted(table, key=operator.itemgetter(col))


def float_eq(a, b):
    """
    Test if two floats are almost equal to each other
    """
    return abs(a - b) < 1e-10 * max(1e-5, abs(a), abs(b))


def linear_interpol(x, x1, x2, y1, y2):
    """
    Linearly interpolate to point x, between
    the points (x1,y1), (x2,y2)
    """
    assert(x1 <= x)
    assert(x2 >= x)
    assert(x1 <= x2)

    alpha = (x - x1) / (x2 - x1)
    return (1. - alpha) * y1 + alpha * y2


def read_table(filename):
    datastream = pkgutil.get_data('burnman', 'data/' + filename)
    datalines = [line.strip()
                 for line in datastream.decode('ascii').split('\n') if line.strip()]
    table = []

    for line in datalines:
        if (line[0] != '#'):
            numbers = np.fromstring(line, sep=' ')
            table.append(numbers)
    return np.array(table)


def array_from_file(filename):
    """
    Generic function to read a file containing floats and commented lines
    into a 2D numpy array.

    Commented lines are prefixed by the characters # or %.
    """
    f = open(filename, 'r')
    data = []
    datastream = f.read()
    f.close()
    datalines = [line.strip().split()
                 for line in datastream.split('\n') if line.strip()]
    for line in datalines:
        if line[0] != "#" and line[0] != "%":
            data.append(map(float, line))

    data = np.array(zip(*data))
    return data


def cut_table(table, min_value, max_value):
    tablen = []
    for i in range(min_value, max_value, 1):
        tablen.append(table[i, :])
    return tablen


def lookup_and_interpolate(table_x, table_y, x_value):
    idx = bisect.bisect_left(table_x, x_value) - 1
    if (idx < 0):
        return table_y[0]
    elif (idx < len(table_x) - 1):
        return linear_interpol(x_value, table_x[idx], table_x[idx + 1],
                               table_y[idx], table_y[idx + 1])
    else:
        return table_y[idx]


def molar_volume_from_unit_cell_volume(unit_cell_v, z):
    """
    Converts a unit cell volume from Angstroms^3 per unitcell,
    to m^3/mol.

    Parameters
    ----------
    unit_cell_v : float
        Unit cell volumes [A^3/unit cell]

    z : float
        Number of formula units per unit cell


    Returns
    -------
    V : float
        Volume [m^3/mol]
    """
    V = unit_cell_v * constants.Avogadro / 1.e30 / z
    return V


def equilibrium_pressure(minerals, stoichiometry, temperature, pressure_initial_guess=1.e5):
    """
    Given a list of minerals, their reaction stoichiometries
    and a temperature of interest, compute the
    equilibrium pressure of the reaction.

    Parameters
    ----------
    minerals : list of minerals
        List of minerals involved in the reaction.

    stoichiometry : list of floats
        Reaction stoichiometry for the minerals provided.
        Reactants and products should have the opposite signs [mol]

    temperature : float
        Temperature of interest [K]

    pressure_initial_guess : optional float
        Initial pressure guess [Pa]

    Returns
    -------
    pressure : float
        The equilibrium pressure of the reaction [Pa]
    """
    def eqm(P, T):
        gibbs = 0.
        for i, mineral in enumerate(minerals):
            mineral.set_state(P[0], T)
            gibbs = gibbs + mineral.gibbs * stoichiometry[i]
        return gibbs

    pressure = fsolve(eqm, [pressure_initial_guess], args=(temperature))[0]

    return pressure


def equilibrium_temperature(minerals, stoichiometry, pressure, temperature_initial_guess=1000.):
    """
    Given a list of minerals, their reaction stoichiometries
    and a pressure of interest, compute the
    equilibrium temperature of the reaction.

    Parameters
    ----------
    minerals : list of minerals
        List of minerals involved in the reaction.

    stoichiometry : list of floats
        Reaction stoichiometry for the minerals provided.
        Reactants and products should have the opposite signs [mol]

    pressure : float
        Pressure of interest [Pa]

    temperature_initial_guess : optional float
        Initial temperature guess [K]

    Returns
    -------
    temperature : float
        The equilibrium temperature of the reaction [K]
    """
    def eqm(T, P):
        gibbs = 0.
        for i, mineral in enumerate(minerals):
            mineral.set_state(P, T[0])
            gibbs = gibbs + mineral.gibbs * stoichiometry[i]
        return gibbs

    temperature = fsolve(eqm, [temperature_initial_guess], args=(pressure))[0]

    return temperature


def invariant_point(minerals_r1, stoichiometry_r1,
                    minerals_r2, stoichiometry_r2,
                    pressure_temperature_initial_guess=[1.e9, 1000.]):
    """
    Given a list of minerals, their reaction stoichiometries
    and a pressure of interest, compute the
    equilibrium temperature of the reaction.

    Parameters
    ----------
    minerals : list of minerals
        List of minerals involved in the reaction.

    stoichiometry : list of floats
        Reaction stoichiometry for the minerals provided.
        Reactants and products should have the opposite signs [mol]

    pressure : float
        Pressure of interest [Pa]

    temperature_initial_guess : optional float
        Initial temperature guess [K]

    Returns
    -------
    temperature : float
        The equilibrium temperature of the reaction [K]
    """
    def eqm(PT):
        P, T = PT
        gibbs_r1 = 0.
        for i, mineral in enumerate(minerals_r1):
            mineral.set_state(P, T)
            gibbs_r1 = gibbs_r1 + mineral.gibbs * stoichiometry_r1[i]
        gibbs_r2 = 0.
        for i, mineral in enumerate(minerals_r2):
            mineral.set_state(P, T)
            gibbs_r2 = gibbs_r2 + mineral.gibbs * stoichiometry_r2[i]
        return [gibbs_r1, gibbs_r2]

    pressure, temperature = fsolve(eqm, pressure_temperature_initial_guess)
    return pressure, temperature


def hugoniot(mineral, P_ref, T_ref, pressures, reference_mineral=None):
    """
    Calculates the temperatures (and volumes) along a Hugoniot
    as a function of pressure according to the Hugoniot equation
    U2-U1 = 0.5*(p2 - p1)(V1 - V2) where U and V are the
    internal energies and volumes (mass or molar) and U = F + TS


    Parameters
    ----------
    mineral : mineral
        Mineral for which the Hugoniot is to be calculated.

    P_ref : float
        Reference pressure [Pa]

    T_ref : float
        Reference temperature [K]

    pressures : numpy array of floats
        Set of pressures [Pa] for which the Hugoniot temperature
        and volume should be calculated

    reference_mineral : mineral
        Mineral which is stable at the reference conditions
        Provides an alternative U_0 and V_0 when the reference
        mineral transforms to the mineral of interest at some
        (unspecified) pressure.

    Returns
    -------
    temperatures : numpy array of floats
        The Hugoniot temperatures at pressure

    volumes : numpy array of floats
        The Hugoniot volumes at pressure
    """

    def Ediff(T, mineral, P, P_ref, U_ref, V_ref):
        mineral.set_state(P, T[0])
        U = mineral.helmholtz + T[0] * mineral.S
        V = mineral.V

        return (U - U_ref) - 0.5 * (P - P_ref) * (V_ref - V)

    if reference_mineral is None:
        reference_mineral = mineral

    reference_mineral.set_state(P_ref, T_ref)
    U_ref = reference_mineral.helmholtz + T_ref * reference_mineral.S
    V_ref = reference_mineral.V

    temperatures = np.empty_like(pressures)
    volumes = np.empty_like(pressures)

    for i, P in enumerate(pressures):
        temperatures[i] = fsolve(
            Ediff, [T_ref], args=(mineral, P, P_ref, U_ref, V_ref))[0]
        volumes[i] = mineral.V

    return temperatures, volumes


def convert_fractions(composite, phase_fractions, input_type, output_type):
    """
    Takes a composite with a set of user defined molar, volume
    or mass fractions (which do not have to be the fractions
    currently associated with the composite) and
    converts the fractions to molar, mass or volume.

    Conversions to and from mass require a molar mass to be
    defined for all phases. Conversions to and from volume
    require set_state to have been called for the composite.

    Parameters
    ----------
    composite : Composite
        Composite for which fractions are to be defined.

    phase_fractions : list of floats
        List of input phase fractions (of type input_type)

    input_type : string
        Input fraction type: 'molar', 'mass' or 'volume'

    output_type : string
        Output fraction type: 'molar', 'mass' or 'volume'

    Returns
    -------
    output_fractions : list of floats
        List of output phase fractions (of type output_type)
    """
    if input_type == 'volume' or output_type == 'volume':
        if composite.temperature is None:
            raise Exception(
                composite.to_string() + ".set_state(P, T) has not been called, so volume fractions are currently undefined. Exiting.")

    if input_type == 'molar':
        molar_fractions = phase_fractions
    if input_type == 'volume':
        total_moles = sum(
            volume_fraction / phase.molar_volume for volume_fraction,
            phase in zip(phase_fractions, composite.phases))
        molar_fractions = [volume_fraction / (phase.molar_volume * total_moles)
                           for volume_fraction, phase in zip(phase_fractions, composite.phases)]
    if input_type == 'mass':
        total_moles = sum(mass_fraction / phase.molar_mass for mass_fraction,
                          phase in zip(phase_fractions, composite.phases))
        molar_fractions = [mass_fraction / (phase.molar_mass * total_moles)
                           for mass_fraction, phase in zip(phase_fractions, composite.phases)]

    if output_type == 'volume':
        total_volume = sum(
            molar_fraction * phase.molar_volume for molar_fraction,
            phase in zip(molar_fractions, composite.phases))
        output_fractions = [molar_fraction * phase.molar_volume
                            / total_volume for molar_fraction, phase in zip(molar_fractions, composite.phases)]
    elif output_type == 'mass':
        total_mass = sum(molar_fraction * phase.molar_mass for molar_fraction,
                         phase in zip(molar_fractions, composite.phases))
        output_fractions = [molar_fraction * phase.molar_mass
                            / total_mass for molar_fraction, phase in zip(molar_fractions, composite.phases)]
    elif output_type == 'molar':
        output_fractions = molar_fractions

    return output_fractions


def bracket(fn, x0, dx, args=(), ratio=1.618, maxiter=100):
    """
    Given a function and a starting guess, find two
    inputs for the function that bracket a root.

    Parameters
    ----------
    fn : function
        The function to bracket
    x0 : float
        The starting guess
    dx : float
        Small step for starting the search
    args : parameter list
        Additional arguments to give to fn
    ratio :
        The step size increases by this ratio
        every step in the search. Defaults to
        the golden ratio.
    maxiter : int
        The maximum number of steps before giving up.

    Returns
    -------
    xa, xb, fa, fb: floats
        xa and xb are the inputs which bracket a root of fn.
        fa and fb are the values of the function at those points.
        If the bracket function takes more than maxiter steps,
        it raises a ValueError.
    """
    niter = 0
    dx = np.abs(dx)
    assert(ratio > 1.0)

    # Get the starting positions
    f0 = fn(x0, *args)
    x_left = x0 - dx
    x_right = x0 + dx
    f_left = fn(x_left, *args)
    f_right = fn(x_right, *args)

    # Overshot zero, try making dx smaller
    if (f0 - f_left) * (f_right - f0) < 0.:
        while (f0 - f_left) * (f_right - f0) < 0. and dx > np.finfo('float').eps and niter < maxiter:
            dx /= ratio
            x_left = x0 - dx
            x_right = x0 + dx
            f_left = fn(x_left, *args)
            f_right = fn(x_right, *args)
            niter += 1
        if niter == maxiter:  # Couldn't find something with same slope in both directions
            raise ValueError('Cannot find zero.')

    niter = 0
    slope = f_right - f0
    if slope > 0. and f0 > 0.:  # Walk left
        dx = -dx
        x1 = x_left
        f1 = f_left
    elif slope > 0. and f0 < 0.:  # Walk right
        x1 = x_right
        f1 = f_right
    elif slope < 0. and f0 > 0:  # Walk right
        x1 = x_right
        f1 = f_right
    else:  # Walk left
        dx = -dx
        x1 = x_left
        f1 = f_left

    # Do the walking
    while f0 * f1 > 0. and niter < maxiter:
        dx *= ratio
        xnew = x1 + dx
        fnew = fn(xnew, *args)
        x0 = x1
        f0 = f1
        x1 = xnew
        f1 = fnew
        niter += 1

    if f0 * f1 > 0.:
        raise ValueError('Cannot find zero.')
    else:
        return x0, x1, f0, f1


def check_eos_consistency(m, P=1.e9, T=300., tol=1.e-4, verbose=False,
                          including_shear_properties=True):
    """
    Compute numerical derivatives of the gibbs free energy of a mineral
    under given conditions, and check these values against those provided
    analytically by the equation of state

    Parameters
    ----------
    m : mineral
        The mineral for which the equation of state
        is to be checked for consistency
    P : float
        The pressure at which to check consistency
    T : float
        The temperature at which to check consistency
    tol : float
        The fractional tolerance for each of the checks
    verbose : boolean
        Decide whether to print information about each
        check
    including_shear_properties : boolean
        Decide whether to check shear information,
        which is pointless for liquids and equations of state
        without shear modulus parameterizations

    Returns
    -------
    consistency: boolean
        If all checks pass, returns True

    """
    dT = 1.
    dP = 1000.

    m.set_state(P, T)
    G0 = m.gibbs
    S0 = m.S
    V0 = m.V

    expr = ['G = F + PV', 'G = H - TS', 'G = E - TS + PV']
    eq = [[m.gibbs, (m.helmholtz + P*m.V)],
          [m.gibbs, (m.H - T*m.S)],
          [m.gibbs, (m.molar_internal_energy - T*m.S + P*m.V)]]

    m.set_state(P, T + dT)
    G1 = m.gibbs
    S1 = m.S
    V1 = m.V

    m.set_state(P + dP, T)
    G2 = m.gibbs
    V2 = m.V

    # T derivatives
    m.set_state(P, T + 0.5*dT)
    expr.extend(['S = -dG/dT', 'alpha = 1/V dV/dT', 'C_p = T dS/dT'])
    eq.extend([[m.S, -(G1 - G0)/dT],
               [m.alpha, (V1 - V0)/dT/m.V],
               [m.molar_heat_capacity_p, (T + 0.5*dT)*(S1 - S0)/dT]])

    # P derivatives
    m.set_state(P + 0.5*dP, T)
    expr.extend(['V = dG/dP', 'K_T = -V dP/dV'])
    eq.extend([[m.V, (G2 - G0)/dP],
               [m.K_T, -0.5*(V2 + V0)*dP/(V2 - V0)]])

    expr.extend(['C_v = Cp - alpha^2*K_T*V*T', 'K_S = K_T*Cp/Cv', 'gr = alpha*K_T*V/Cv'])
    eq.extend([[m.molar_heat_capacity_v, m.molar_heat_capacity_p - m.alpha*m.alpha*m.K_T*m.V*T],
               [m.K_S, m.K_T*m.molar_heat_capacity_p/m.molar_heat_capacity_v],
               [m.gr, m.alpha*m.K_T*m.V/m.molar_heat_capacity_v]])

    expr.append('Vphi = np.sqrt(K_S/rho)')
    eq.append([m.bulk_sound_velocity, np.sqrt(m.K_S/m.rho)])

    if including_shear_properties:
        expr.extend(['Vp = np.sqrt((K_S + 4G/3)/rho)', 'Vs = np.sqrt(G_S/rho)'])

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            eq.extend([[m.p_wave_velocity, np.sqrt((m.K_S + 4.*m.G/3.)/m.rho)],
                       [m.shear_wave_velocity, np.sqrt(m.G/m.rho)]])
            if len(w) == 1:
                print(w[0].message)
                print('\nYou can suppress this message by setting the '
                      'parameter\nincluding_shear_properties to False '
                      'when calling check_eos_consistency.\n')
        note = ''
    else:
        note = ' (not including shear properties)'

    consistencies = [np.abs(e[0] - e[1]) < np.abs(tol*e[1]) + np.finfo('float').eps for e in eq]
    eos_is_consistent = np.all(consistencies)

    if verbose:
        print('Checking EoS consistency for {0:s}{1}'.format(m.to_string(), note))
        print('Expressions within tolerance of {0:2f}'.format(tol))
        for i, c in enumerate(consistencies):
            print('{0:10s} : {1:5s}'.format(expr[i], str(c)))
        if eos_is_consistent:
            print('All EoS consistency constraints satisfied for {0:s}'.format(m.to_string()))
        else:
            print('Not satisfied all EoS consistency constraints for {0:s}'.format(m.to_string()))

    return eos_is_consistent


def _pad_ndarray_inverse_mirror(array, padding):
    """
    Pads an ndarray according to an inverse mirror
    scheme. For example, for a 1D array
    [2, 4, 6, 7, 8] padded by 3 cells, we have:

     padding  |  original array |  padding

    -3 -2  0  |  2  4  6  7  8  |  9 10 12

    Parameters
    ----------
    array : numpy ndarray
        The array to be padded
    padding : tuple
        The number of elements with which to pad the
        array in each dimension.

    Returns
    -------
    padded_array: numpy ndarray
        The padded array

    """
    padded_shape = [n + 2*padding[i] for i, n in enumerate(array.shape)]
    padded_array = np.zeros(padded_shape)

    slices = tuple([slice(padding[i], padding[i] + l) for i, l in enumerate(array.shape)])
    padded_array[slices] = array

    padded_array_indices = list(itertools.product(*[range(n + 2*padding[i]) for i, n in enumerate(array.shape)]))
    inserted_indices = list(itertools.product(*[range(padding[i], padding[i] + l) for i, l in enumerate(array.shape)]))
    padded_array_indices.extend(inserted_indices)

    counter = Counter(padded_array_indices)
    keys = list(counter.keys())
    padded_indices = [keys[i] for i, value in enumerate(counter.values()) if value == 1]
    edge_indices = tuple([tuple([np.min([np.max([axis_idx, padding[dimension]]), padded_array.shape[dimension] - padding[dimension] - 1])
                                 for dimension, axis_idx in enumerate(idx)]) for idx in padded_indices])
    mirror_indices = tuple([tuple([2*edge_indices[i][j] - padded_indices[i][j] for j in range(len(array.shape))]) for i in range(len(padded_indices))])

    for i, idx in enumerate(padded_indices):
        padded_array[idx] = 2.*padded_array[edge_indices[i]] - padded_array[mirror_indices[i]]

    return padded_array


def smooth_array(array, grid_spacing,
                 gaussian_rms_widths, truncate=4.0,
                 mode='inverse_mirror'):
    """
    Creates a smoothed array by convolving it with a gaussian filter.
    Grid resolutions and gaussian RMS widths are required for each of
    the axes of the numpy array. The smoothing is truncated at a
    user-defined number of standard deviations. The edges of the array
    can be padded in a number of different ways given by the
    'mode' parameter.

    Parameters
    ----------
    array : numpy ndarray
        The array to smooth
    grid_spacing : numpy array of floats
        The spacing of points along each axis
    gaussian_rms_widths : numpy array of floats
        The Gaussian RMS widths/standard deviations for the
        Gaussian convolution.
    truncate : float (default=4.)
        The number of standard deviations at which to truncate
        the smoothing.
    mode : {'reflect', 'constant', 'nearest', 'mirror', 'wrap', 'inverse_mirror'}
        The mode parameter determines how the array borders are handled
        either by scipy.ndimage.filters.gaussian_filter.
        Default is 'inverse_mirror', which uses
        burnman.tools._pad_ndarray_inverse_mirror().

    Returns
    -------
    smoothed_array: numpy ndarray
       The smoothed array

    """

    # gaussian_filter works with standard deviations normalised to
    # the grid spacing.
    sigma = tuple(np.array(gaussian_rms_widths)/np.array(grid_spacing))

    if mode == 'inverse_mirror':
        padding = tuple([int(np.ceil(truncate*s)) for s in sigma])
        padded_array = _pad_ndarray_inverse_mirror(array, padding)
        smoothed_padded_array = gaussian_filter(padded_array,
                                                sigma=sigma)
        slices = tuple([slice(padding[i], padding[i] + l) for i, l in enumerate(array.shape)])
        smoothed_array = smoothed_padded_array[slices]
    else:
        smoothed_array = gaussian_filter(array, sigma=sigma, mode=mode)

    return smoothed_array


def interp_smoothed_array_and_derivatives(array,
                                          x_values, y_values,
                                          x_stdev=0., y_stdev=0.,
                                          truncate=4.,
                                          mode='inverse_mirror',
                                          indexing='xy'):
    """
    Creates a smoothed array on a regular 2D grid. Smoothing
    is achieved using burnman.tools.smooth_array().
    Outputs scipy.interpolate.interp2d() interpolators
    which can be used to query the array, or its derivatives in the
    x- and y- directions.

    Parameters
    ----------
    array : 2D numpy array
        The array to smooth. Each element array[i][j]
        corresponds to the position x_values[i], y_values[j]
    x_values : 1D numpy array
        The gridded x values over which to create the smoothed grid
    y_values : 1D numpy array
        The gridded y_values over which to create the smoothed grid
    x_stdev : float
        The standard deviation for the Gaussian filter along the x axis
    y_stdev : float
        The standard deviation for the Gaussian filter along the x axis
    truncate : float (optional)
        The number of standard deviations at which to truncate
        the smoothing (default = 4.).
    mode : {'reflect', 'constant', 'nearest', 'mirror', 'wrap', 'inverse_mirror'}
        The mode parameter determines how the array borders are handled
        either by scipy.ndimage.filters.gaussian_filter.
        Default is 'inverse_mirror', which uses
        burnman.tools._pad_ndarray_inverse_mirror().
    indexing : {'xy', 'ij'}, optional
        Cartesian ('xy', default) or matrix ('ij') indexing of output.
        See numpy.meshgrid for more details.

    Returns
    -------
    interps: tuple of three interp2d functors
        interpolation functions for the smoothed property and
        the first derivatives with respect to x and y.

    """

    dx = x_values[1] - x_values[0]
    dy = y_values[1] - y_values[0]

    if indexing == 'xy':
        smoothed_array = smooth_array(array=array,
                                      grid_spacing=np.array([dy, dx]),
                                      gaussian_rms_widths=np.array([y_stdev, x_stdev]),
                                      truncate=truncate,
                                      mode=mode)

    elif indexing == 'ij':
        smoothed_array = smooth_array(array=array,
                                      grid_spacing=np.array([dx, dy]),
                                      gaussian_rms_widths=np.array([x_stdev, y_stdev]),
                                      truncate=truncate,
                                      mode=mode).T

    else:
        raise Exception('Indexing scheme not recognised. Should be ij or xy.')

    dSAdydy, dSAdxdx = np.gradient(smoothed_array)

    interps = (interp2d(x_values, y_values, smoothed_array, kind='linear'),
               interp2d(x_values, y_values, dSAdxdx/dx, kind='linear'),
               interp2d(x_values, y_values, dSAdydy/dy, kind='linear'))

    return interps


def attribute_function(m, attributes, powers=[]):
    """
    Function which returns a function which can be used to
    evaluate material properties at a point. This function
    allows the user to define the property returned
    as a string. The function can itself be passed to another
    function
    (such as nonlinear_fitting.confidence_prediction_bands()).

    Properties can either be simple attributes (e.g. K_T) or
    a product of attributes, each raised to some power.

    Parameters
    ----------
    m : Material
        The material instance evaluated by the output function.
    attributes : list of strings
        The list of material attributes / properties to
        be evaluated in the product
    powers : list of floats
        The powers to which each attribute should be raised
        during evaluation
    Returns
    -------
    f : function(x)
        Function which returns the value of product(a_i**p_i)
        as a function of condition (x = [P, T, V])
    """
    if type(attributes) is str:
        attributes = [attributes]
    if powers == []:
        powers = [1. for a in attributes]

    def f(x):
        P, T, V = x
        m.set_state(P, T)
        value = 1.
        for a, p in zip(*[attributes, powers]):
            value *= np.power(getattr(m, a), p)
        return value
    return f


def compare_l2(depth, calc, obs):
    """
    Computes the L2 norm for N profiles at a time (assumed to be linear between points).

    :type depths: array of float
    :param depths: depths. :math:`[m]`
    :type calc: list of arrays of float
    :param calc: N arrays calculated values, e.g. [mat_vs,mat_vphi]
    :type obs: list of arrays of float
    :param obs: N arrays of values (observed or calculated) to compare to , e.g. [seis_vs, seis_vphi]

    :returns: array of L2 norms of length N
    :rtype: array of floats
    """
    err = []
    for i in range(len(calc)):
        err.append(l2(depth, calc[i], obs[i]))

    return err


def compare_chifactor(calc, obs):
    """
    Computes the chi factor for N profiles at a time. Assumes a 1% a priori uncertainty on the seismic model.


    :type calc: list of arrays of float
    :param calc: N arrays calculated values, e.g. [mat_vs,mat_vphi]
    :type obs: list of arrays of float
    :param obs: N arrays of values (observed or calculated) to compare to , e.g. [seis_vs, seis_vphi]

    :returns: error array of length N
    :rtype: array of floats
    """
    err = []
    for i in range(len(calc)):
        err.append(chi_factor(calc[i], obs[i]))

    return err


def l2(x, funca, funcb):
    """
    Computes the L2 norm for one profile(assumed to be linear between points).

    :type x: array of float
    :param x: depths :math:`[m]`.
    :type funca: list of arrays of float
    :param funca: array calculated values
    :type funcb: list of arrays of float
    :param funcb: array of values (observed or calculated) to compare to

    :returns: L2 norm
    :rtype: array of floats
    """
    diff = np.array(funca - funcb)
    diff = diff * diff

    return integrate.trapz(diff, x)


def nrmse(x, funca, funcb):
    """
    Normalized root mean square error for one profile
    :type x: array of float
    :param x: depths in m.
    :type funca: list of arrays of float
    :param funca: array calculated values
    :type funcb: list of arrays of float
    :param funcb: array of values (observed or calculated) to compare to

    :returns: RMS error
    :rtype: array of floats

    """
    diff = np.array(funca - funcb)
    diff = diff * diff
    rmse = np.sqrt(np.sum(diff) / x)
    nrmse = rmse / (np.max(funca) - np.min(funca))

    return nrmse


def chi_factor(calc, obs):
    """
    :math:`\\chi` factor for one profile assuming 1% uncertainty on the reference model (obs)
    :type calc: list of arrays of float
    :param calc: array calculated values
    :type obs: list of arrays of float
    :param obs: array of reference values to compare to

    :returns: :math:`\\chi` factor
    :rtype: array of floats

    """

    err = np.empty_like(calc)
    for i in range(len(calc)):
        err[i] = pow((calc[i] - obs[i]) / (0.01 * np.mean(obs)), 2.)

    err_tot = np.sum(err) / len(err)

    return err_tot



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import
import numpy as np
from scipy.linalg import lu

from .processchemistry import *
from . import constants
from . import solidsolution

# This module computes chemical potentials (partial molar gibbs free
# energies) for an assemblage based on the Gibbs free energies and
# compositions of the individual phases.

# It can also calculate fugacities based on the gibbs free energies of the
# endmembers corresponding to chemical components.

# It can also calculate fugacities relative to other bulk compositions


def chemical_potentials(assemblage, component_formulae):
    """
    The compositional space of the components does not have to be a
    superset of the compositional space of the assemblage. Nor do they have to
    compose an orthogonal basis.

    The components must each be described by a linear mineral combination

    The mineral compositions must be linearly independent

        Parameters
        ----------
        assemblage : list of classes
            List of material classes
            set_method and set_state should already have been used
            the composition of the solid solutions should also have been set

        component_formulae : list of dictionaries
            List of chemical component formula dictionaries
            No restriction on length

        Returns
        -------
        component_potentials : array of floats
            Array of chemical potentials of components

    """
    # Split solid solutions into their respective endmembers
    # Find the chemical potentials of all the endmembers
    endmember_list = []
    endmember_potentials = []
    for mineral in assemblage:
        if isinstance(mineral, solidsolution.SolidSolution):
            for member in mineral.endmembers:
                endmember_list.append(member[0])
            for potential in mineral.partial_gibbs:
                endmember_potentials.append(potential)
        else:
            endmember_list.append(mineral)
            endmember_potentials.append(mineral.gibbs)

    # Make an array of all the endmember formulae
    endmember_formulae = [endmember.params['formula']
                          for endmember in endmember_list]
    endmember_compositions, elements = compositional_array(endmember_formulae)

    pl, u = lu(endmember_compositions, permute_l=True)
    assert(min(np.dot(np.square(u), np.ones(shape=len(elements)))) > 1e-6), \
        'Endmember compositions do not form an independent set of basis vectors'

    # Make an array of component formulae with elements in the same order as
    # the endmember array
    component_compositions = ordered_compositional_array(
        component_formulae, elements)

    p = np.linalg.lstsq(endmember_compositions.T, component_compositions.T, rcond=-1)
    # rcond=-1 is old default to silence warning and be compatible with python 2.7
    for idx, error in enumerate(p[1]):
        assert (error < 1e-10), \
            'Component %d not defined by prescribed assemblage' % (idx + 1)

    # Create an array of endmember proportions which sum to each component
    # composition
    endmember_proportions = np.around(p[0], 10).T

    # Calculate the chemical potential of each component
    component_potentials = np.dot(endmember_proportions, endmember_potentials)
    return component_potentials


def fugacity(standard_material, assemblage):
    """
        Parameters
        ----------
        standard_material: class
            Material class
            set_method and set_state should already have been used
            material must have a formula as a dictionary parameter

        assemblage: list of classes
            List of material classes
            set_method and set_state should already have been used

        Returns
        -------
        fugacity : float
            Value of the fugacity of the component with respect to
            the standard material

    """
    component_formula = standard_material.params['formula']
    chemical_potential = chemical_potentials(
        assemblage, [component_formula])[0]

    fugacity = np.exp((chemical_potential - standard_material.gibbs) / (
        constants.gas_constant * assemblage[0].temperature))
    return fugacity


def relative_fugacity(standard_material, assemblage, reference_assemblage):
    """
        Parameters
        ----------
        standard_material: class
            Material class
            set_method and set_state should already have been used
            material must have a formula as a dictionary parameter

        assemblage: list of classes
            List of material classes
            set_method and set_state should already have been used

        reference_assemblage: list of classes
            List of material classes
            set_method and set_state should already have been used

        Returns
        -------
        relative_fugacity : float
            Value of the fugacity of the component in the assemblage
            with respect to the reference_assemblage

    """
    component_formula = standard_material.params['formula']
    chemical_potential = chemical_potentials(
        assemblage, [component_formula])[0]
    reference_chemical_potential = chemical_potentials(
        reference_assemblage, [component_formula])[0]

    relative_fugacity = np.exp((chemical_potential - reference_chemical_potential) / (
        constants.gas_constant * assemblage[0].temperature))
    return relative_fugacity



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import
from __future__ import print_function

import numpy as np
import matplotlib.pyplot as plt

from .tools import unit_normalize
from .material import Material, material_property

class AnisotropicMaterial(Material):
    """
    A class that represents an anisotropic elastic material. This class
    is initialised with a set of elastic constants and a density. It can
    then be interrogated to find the values of different properties,
    such as bounds on seismic velocities. There are also several functions
    which can be called to calculate properties along directions oriented
    with respect to the elastic tensor. Initialization is via a density
    and a full stiffness tensor in Voigt notation

    See :cite:`Mainprice2011` Geological Society of London Special Publication
    and https://materialsproject.org/wiki/index.php/Elasticity_calculations
    for mathematical descriptions of each function.
    """

    def __init__(self, rho, cijs):
        self.params = {'rho_0': rho,
                       'stiffness_tensor_0': cijs}

        assert self.params['stiffness_tensor_0'].shape == (6, 6), 'stiffness_tensor must be in Voigt notation (6x6)'
        assert np.allclose(self.params['stiffness_tensor_0'].T,
                           self.params['stiffness_tensor_0']), 'stiffness_tensor must be symmetric'

        Material.__init__(self)

    def _voigt_index_to_ij(self, m):
        """
        Returns the ij (or kl) indices of the
        stiffness tensor which correspond to those
        of the Voigt notation m (or n).
        """
        if m == 3:
            return 1, 2
        elif m == 4:
            return 0, 2
        elif m == 5:
            return 0, 1
        else:
            return m, m

    def _voigt_notation_to_stiffness_tensor(self, voigt_notation):
        """
        Converts a stiffness tensor in Voigt notation (6x6 matrix)
        to the full fourth rank tensor (3x3x3x3 matrix).
        """
        stiffness_tensor = np.zeros([3, 3, 3, 3])
        for m in range(6):
            i, j = self._voigt_index_to_ij(m)
            for n in range(6):
                k, l = self._voigt_index_to_ij(n)
                stiffness_tensor[i][j][k][l] = voigt_notation[m][n]
                stiffness_tensor[j][i][k][l] = voigt_notation[m][n]
                stiffness_tensor[i][j][l][k] = voigt_notation[m][n]
                stiffness_tensor[j][i][l][k] = voigt_notation[m][n]
        return stiffness_tensor


    @material_property
    def stiffness_tensor(self):
        return self.params['stiffness_tensor_0']

    @material_property
    def full_stiffness_tensor(self):
        return self._voigt_notation_to_stiffness_tensor(self.stiffness_tensor)

    @material_property
    def compliance_tensor(self):
        return np.linalg.inv(self.stiffness_tensor)

    @material_property
    def full_compliance_tensor(self):
        try: # numpy.block was new in numpy version 1.13.0.
            block = np.block([[ np.ones((3, 3)), 2.*np.ones((3, 3))],
                              [2.*np.ones((3, 3)), 4.*np.ones((3, 3))]])
        except:
            block = np.array(np.bmat( [[[[1.]*3]*3, [[2.]*3]*3], [[[2.]*3]*3, [[4.]*3]*3]] ))
        return self._voigt_notation_to_stiffness_tensor(np.divide(self.compliance_tensor, block))

    @material_property
    def density(self):
        return self.params['rho_0']

    @material_property
    def bulk_modulus_voigt(self):
        """
        Computes the bulk modulus (Voigt bound)
        """
        K = np.sum([[self.stiffness_tensor[i][k] for k in range(3)] for i in range(3)])/9.
        return K

    @material_property
    def bulk_modulus_reuss(self):
        """
        Computes the bulk modulus (Reuss bound)
        """
        beta = np.sum([[self.compliance_tensor[i][k] for k in range(3)] for i in range(3)])
        return 1./beta

    @material_property
    def bulk_modulus_vrh(self):
        """
        Computes the bulk modulus (Voigt-Reuss-Hill average)
        """
        return 0.5*(self.bulk_modulus_voigt + self.bulk_modulus_reuss)

    @material_property
    def shear_modulus_voigt(self):
        """
        Computes the shear modulus (Voigt bound)
        """
        G = ( np.sum([self.stiffness_tensor[i][i] for i in [0, 1, 2]]) +
              np.sum([self.stiffness_tensor[i][i] for i in [3, 4, 5]])*3. -
              ( self.stiffness_tensor[0][1] +
                self.stiffness_tensor[1][2] +
                self.stiffness_tensor[2][0] )) / 15.
        return G

    @material_property
    def shear_modulus_reuss(self):
        """
        Computes the shear modulus (Reuss bound)
        """
        beta =  ( np.sum([self.compliance_tensor[i][i] for i in [0, 1, 2]])*4. +
                  np.sum([self.compliance_tensor[i][i] for i in [3, 4, 5]])*3. -
                  ( self.compliance_tensor[0][1] +
                    self.compliance_tensor[1][2] +
                    self.compliance_tensor[2][0])*4. ) / 15.
        return 1./beta

    @material_property
    def shear_modulus_vrh(self):
        """
        Computes the shear modulus (Voigt-Reuss-Hill average)
        """
        return 0.5*(self.shear_modulus_voigt + self.shear_modulus_reuss)

    @material_property
    def universal_elastic_anisotropy(self):
        """
        Compute the universal elastic anisotropy
        """
        return ( 5.*(self.shear_modulus_voigt/self.shear_modulus_reuss) +
                 (self.bulk_modulus_voigt/self.bulk_modulus_reuss) - 6. )

    @material_property
    def isotropic_poisson_ratio(self):
        """
        Compute mu, the isotropic Poisson ratio
        (a description of the laterial response to loading)
        """
        return ( (3.*self.bulk_modulus_vrh - 2.*self.shear_modulus_vrh) /
                 (6.*self.bulk_modulus_vrh + 2.*self.shear_modulus_vrh) )

    def christoffel_tensor(self, propagation_direction):
        """
        Computes the Christoffel tensor from an elastic stiffness
        tensor and a propagation direction for a seismic wave
        relative to the stiffness tensor

        T_ik = C_ijkl n_j n_l
        """
        propagation_direction = unit_normalize(propagation_direction)
        Tik = np.tensordot(np.tensordot(self.full_stiffness_tensor,
                                        propagation_direction,
                                        axes=([1],[0])),
                           propagation_direction,
                           axes=([2],[0]))
        return Tik

    def linear_compressibility(self, direction):
        """
        Computes the linear compressibility in a given direction
        relative to the stiffness tensor
        """
        direction = unit_normalize(direction)
        Sijkk = np.einsum('ijkk', self.full_compliance_tensor)
        beta = Sijkk.dot(direction).dot(direction)
        return beta

    def youngs_modulus(self, direction):
        """
        Computes the Youngs modulus in a given direction
        relative to the stiffness tensor
        """
        direction = unit_normalize(direction)
        Sijkl = self.full_compliance_tensor
        S = Sijkl.dot(direction).dot(direction).dot(direction).dot(direction)
        return 1./S

    def shear_modulus(self, plane_normal, shear_direction):
        """
        Computes the shear modulus on a plane in a given
        shear direction relative to the stiffness tensor
        """
        plane_normal = unit_normalize(plane_normal)
        shear_direction = unit_normalize(shear_direction)

        assert np.abs(plane_normal.dot(shear_direction)) < np.finfo(float).eps, 'plane_normal and shear_direction must be orthogonal'
        Sijkl = self.full_compliance_tensor
        G = Sijkl.dot(shear_direction).dot(plane_normal).dot(shear_direction).dot(plane_normal)
        return 0.25/G

    def poissons_ratio(self,
                       axial_direction,
                       lateral_direction):
        """
        Computes the poisson ratio given loading and response
        directions relative to the stiffness tensor
        """

        axial_direction = unit_normalize(axial_direction)
        lateral_direction = unit_normalize(lateral_direction)
        assert np.abs(axial_direction.dot(lateral_direction)) < np.finfo(float).eps, 'axial_direction and lateral_direction must be orthogonal'

        Sijkl = self.full_compliance_tensor
        x = axial_direction
        y = lateral_direction
        nu = -(Sijkl.dot(y).dot(y).dot(x).dot(x) /
               Sijkl.dot(x).dot(x).dot(x).dot(x) )
        return nu

    def wave_velocities(self, propagation_direction):
        """
        Computes the compressional wave velocity, and two
        shear wave velocities in a given propagation direction

        Returns two lists, containing the wave speeds and
        directions of particle motion relative to the stiffness tensor
        """
        propagation_direction = unit_normalize(propagation_direction)

        Tik = self.christoffel_tensor(propagation_direction)

        eigenvalues, eigenvectors = np.linalg.eig(Tik)

        idx = eigenvalues.argsort()[::-1]
        eigenvalues = np.real(eigenvalues[idx])
        eigenvectors = eigenvectors[:,idx]
        velocities = np.sqrt(eigenvalues/self.rho)

        return velocities, eigenvectors

def voigt_array_from_cijs(cijs, index_lists):
    C = np.zeros([6, 6])
    for i, index_list in enumerate(index_lists):
        for indices in index_list:
            C[indices] = cijs[i]
            C[indices[::-1]] = cijs[i]
    return C

class IsotropicMaterial(AnisotropicMaterial):
    """
    A class derived from the AnisotropicMaterial base class
    Initialization takes two input parameters; rho and
    [C12, C44] (i.e. lambda and mu, the Lame parameters)
    """
    def __init__(self, rho, cijs):

        assert len(cijs) == 2
        cijs = list(cijs)
        cijs.insert(0, cijs[0] + 2.*cijs[1]) # C11 = C12 + 2C44
        index_lists = [[(0, 0), (1, 1), (2, 2)], # C11
                       [(0, 1), (0, 2), (1, 2)], # C12
                       [(3, 3), (4, 4), (5, 5)]] # C44

        AnisotropicMaterial.__init__(
            self, rho, voigt_array_from_cijs(cijs, index_lists))

class CubicMaterial(AnisotropicMaterial):
    """
    A class derived from the AnisotropicMaterial base class
    Initialization takes two input parameters; rho and
    [C11, C12, C44]
    """
    def __init__(self, rho, cijs):

        assert len(cijs) == 3
        index_lists = [[(0, 0), (1, 1), (2, 2)], # C11
                       [(0, 1), (0, 2), (1, 2)], # C12
                       [(3, 3), (4, 4), (5, 5)]] # C44

        AnisotropicMaterial.__init__(
            self, rho, voigt_array_from_cijs(cijs, index_lists))

class HexagonalMaterial(AnisotropicMaterial):
    """
    A class derived from the AnisotropicMaterial base class
    Initialization takes two input parameters; rho and
    [C11, C12, C13, C33, C44]
    """
    def __init__(self, rho, cijs):
        assert len(cijs) == 5
        cijs = list(cijs)
        cijs.append((cijs[0] - cijs[1])/2.) # C66 = (C11-C12)/2.

        index_lists = [[(0, 0), (1, 1)], # C11
                       [(0, 1)], # C12
                       [(0, 2), (1, 2)], # C13
                       [(2, 2)], # C33
                       [(3, 3), (4, 4)], # C44
                       [(5, 5)]] # C66

        AnisotropicMaterial.__init__(
            self, rho, voigt_array_from_cijs(cijs, index_lists))

class TetragonalMaterial(AnisotropicMaterial):
    """
    A class derived from the AnisotropicMaterial base class
    Initialization takes two input parameters; rho and
    [C11, C12, C13, C33, C44, C66] or
    [C11, C12, C13, C16, C33, C44, C66]
    """
    def __init__(self, rho, cijs):
        if len(cijs) == 6:
            # Tetragonal I / Laue class 4/mmm
            index_lists = [[(0, 0), (1, 1)], # C11
                           [(0, 1)], # C12
                           [(0, 2), (1, 2)], # C13
                           [(2, 2)], # C33
                           [(3, 3), (4, 4)], # C44
                           [(5, 5)]] # C66
        elif len(cijs) == 7:
            # Tetragonal II / Laue class 4/m
            cijs = list(cijs)
            cijs.insert(4, -cijs[3]) # C26 = -C16
            index_lists = [[(0, 0), (1, 1)], # C11
                           [(0, 1)], # C12
                           [(0, 2), (1, 2)], # C13
                           [(0, 5)], # C16
                           [(1, 5)], # C26
                           [(2, 2)], # C33
                           [(3, 3), (4, 4)], # C44
                           [(5, 5)]] # C66
        else:
            raise Exception('Tetragonal materials should have '
                            'either 6 or 7 independent Cijs')

        AnisotropicMaterial.__init__(
            self, rho, voigt_array_from_cijs(cijs, index_lists))


class RhombohedralMaterial(AnisotropicMaterial):
    """
    A class derived from the AnisotropicMaterial base class
    Initialization takes two input parameters; rho and
    [C11, C12, C13, C14, C33, C44, C66] or
    [C11, C12, C13, C14, C15, C33, C44, C66]
    """
    def __init__(self, rho, cijs):
        cijs = list(cijs)
        if len(cijs) == 7:
            # Rhombohedral I / Laue class \bar{3}m
            cijs.insert(4, -cijs[3]) # C24 = -C14
            index_lists = [[(0, 0), (1, 1)], # C11
                           [(0, 1)], # C12
                           [(0, 2), (1, 2)], # C13
                           [(0, 3), (4, 5)], # C14
                           [(1, 3)], # C24
                           [(2, 2)], # C33
                           [(3, 3), (4, 4)], # C44
                           [(5, 5)]] # C66

        elif len(cijs) == 8:
            # Rhombohedral II / Laue class \bar{3}
            cijs.insert(4, -cijs[3]) # C24 = -C14
            cijs.insert(6, -cijs[5]) # C25 = -C15
            index_lists = [[(0, 0), (1, 1)], # C11
                           [(0, 1)], # C12
                           [(0, 2), (1, 2)], # C13
                           [(0, 3), (4, 5)], # C14
                           [(1, 3)], # C24
                           [(0, 4)], # C15
                           [(1, 4), (3, 5)], # C25
                           [(2, 2)], # C33
                           [(3, 3), (4, 4)], # C44
                           [(5, 5)]] # C66

        else:
            raise Exception('Rhombohedral materials should have '
                            'either 7 or 8 independent Cijs')

        AnisotropicMaterial.__init__(
            self, rho, voigt_array_from_cijs(cijs, index_lists))

class OrthorhombicMaterial(AnisotropicMaterial):
    """
    A class derived from the AnisotropicMaterial base class
    Initialization takes two input parameters; rho and
    [C11, C12, C13, C22, C23, C33, C44, C55, C66]
    """
    def __init__(self, rho, cijs):

        assert len(cijs) == 9
        index_lists = [[(0, 0)], # C11
                       [(0, 1)], # C12
                       [(0, 2)], # C13
                       [(1, 1)], # C22
                       [(1, 2)], # C23
                       [(2, 2)], # C33
                       [(3, 3)], # C44
                       [(4, 4)], # C55
                       [(5, 5)]] # C66

        AnisotropicMaterial.__init__(
            self, rho, voigt_array_from_cijs(cijs, index_lists))


class MonoclinicMaterial(AnisotropicMaterial):
    """
    A class derived from the AnisotropicMaterial base class
    Initialization takes two input parameters; rho and
    [C11, C12, C13, C15, C22, C23, C25, C33, C35, C44, C46, C55, C66]
    """
    def __init__(self, rho, cijs):

        assert len(cijs) == 13
        index_lists = [[(0, 0)], # C11
                       [(0, 1)], # C12
                       [(0, 2)], # C13
                       [(0, 4)], # C15
                       [(1, 1)], # C22
                       [(1, 2)], # C23
                       [(1, 4)], # C25
                       [(2, 2)], # C33
                       [(2, 4)], # C35
                       [(3, 3)], # C44
                       [(3, 5)], # C46
                       [(4, 4)], # C55
                       [(5, 5)]] # C66

        AnisotropicMaterial.__init__(
            self, rho, voigt_array_from_cijs(cijs, index_lists))

class TriclinicMaterial(AnisotropicMaterial):
    """
    A class derived from the AnisotropicMaterial base class
    Initialization takes two input parameters; rho and
    [Cij, where 1<=i<=6 and i<=j<=6]
    """
    def __init__(self, rho, cijs):

        assert len(cijs) == 21
        index_lists=[[(i, j)] for i in range(6) for j in range(i, 6)]

        AnisotropicMaterial.__init__(
            self, rho, voigt_array_from_cijs(cijs, index_lists))



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2019 by the BurnMan team, released under the GNU
# GPL v2 or later.


# This is a standalone program that converts the Holland and Powell data format
# into the standard burnman format (printed to file)
# It only outputs properties of solid endmembers - other endmembers are
# currently ignored.


import sys
import os.path
import pprint
from collections import OrderedDict
# hack to allow scripts to be placed in subdirectories next to burnman:
if not os.path.exists('burnman') and os.path.exists('../../../burnman'):
    sys.path.insert(1, os.path.abspath('../../..'))

import burnman
from burnman.processchemistry import dictionarize_formula, formula_mass


if os.path.isfile('tc-ds633.txt') == False:
    print('This code requires the data file tc-ds633.txt.')
    print(
        'This file is bundled with the software package THERMOCALC, which can be found here:')
    print(
        'https://www.esc.cam.ac.uk/research/research-groups/research-projects/tim-hollands-software-pages/thermocalc')
    print('')
    print('Please download the file and place it in this directory.')
    exit()

# Components
components = ['Si', 'Ti', 'Al', 'Fe', 'Mg', 'Mn', 'Ca', 'Na',
              'K', 'O', 'H', 'C', 'Cl', 'e-', 'Ni', 'Zr', 'S', 'Cu', 'Cr']


class Endmember:

    def __init__(self, name, atoms, formula, sites, comp, H, S, V, Cp, a, k, flag, od):
        if flag != -1 and flag != -2 and k[0] > 0:
            formula = dictionarize_formula(formula)
            self.params = OrderedDict([('name', name),
                                       ('formula', formula),
                                       ('equation_of_state', 'hp_tmt'),
                                       ('H_0', round(H * 1e3, 10)),
                                       ('S_0', round(S * 1e3, 10)),
                                       ('V_0', round(V * 1e-5, 15)),
                                       ('Cp', [round(Cp[0] * 1e3, 10),
                                               round(Cp[1] * 1e3, 10),
                                               round(Cp[2] * 1e3, 10),
                                               round(Cp[3] * 1e3, 10)]),
                                       ('a_0', a),
                                       ('K_0', round(k[0] * 1e8, 10)),
                                       ('Kprime_0', k[1]),
                                       ('Kdprime_0', round(k[2] * 1e-8, 15)),
                                       ('n', sum(formula.values())),
                                       ('molar_mass', round(formula_mass(formula), 10))])
        if flag == 1:
            self.landau_hp = OrderedDict([('P_0', 1e5),
                                          ('T_0', 298.15),
                                          ('Tc_0', od[0]),
                                          ('S_D', round(od[1] * 1e3, 10)),
                                          ('V_D', round(od[2] * 1e-5, 10))])

        if flag == 2:
            self.bragg_williams = OrderedDict([('deltaH', round(od[0] * 1e3, 10)),
                                               ('deltaV', round(od[1] * 1e-5, 15)),
                                               ('Wh', round(od[2] * 1e3, 10)),
                                               ('Wv', round(od[3] * 1e-5, 15)),
                                               ('n', od[4]),
                                               ('factor', od[5])])




# Read dataset
with open('tc-ds633.txt', 'r') as f:
    ds = [line.split() for line in f]

def getmbr(ds, mbr):
    mbrarray = []
    for i in range(0, int(ds[0][0])):
        if ds[i * 4 + 3][0] == mbr:
            atoms = 0.0
            formula = ''
            for j in range(3, len(ds[i * 4 + 3]) - 1, 2):
                atoms += float(ds[i * 4 + 3][j])
                formula = formula + \
                    components[int(ds[i * 4 + 3][j - 1]) - 1] + str(
                        round(float(ds[i * 4 + 3][j]), 10))
            if mbr.endswith('L'):
                flag = -2
                od = [0]
            else:
                flag = int(ds[i * 4 + 6][4])
            endmember = Endmember(mbr, atoms, formula, int(ds[i * 4 + 3][1]),
                                  list(map(float, ds[i * 4 + 3][2:(len(ds[i * 4 + 3]) - 1)])),
                                  float(ds[i * 4 + 4][0]), float(ds[i * 4 + 4][1]),
                                  float(ds[i * 4 + 4][2]), map(float, ds[i * 4 + 5]),
                                  float(ds[i * 4 + 6][0]), list(map(float, ds[i * 4 + 6][1:4])),
                                  flag, list(map(float, ds[i * 4 + 6][5:])))
            return endmember

with open('HGP_2018_ds633.py', 'wb') as outfile:
    outfile.write('# This file is part of BurnMan - a thermoelastic and '
                  'thermodynamic toolkit for the Earth and Planetary Sciences\n'
                  '# Copyright (C) 2012 - 2019 by the BurnMan team, '
                  'released under the GNU \n# GPL v2 or later.\n\n\n'
                  '"""\n'
                  'HGP_2018 (ds-62)\n'
                  'Endmember minerals from Holland, Green and Powell (2018) and references therein\n'
                  'Dataset version 6.33\n'
                  'The values in this document are all in S.I. units,\n'
                  'unlike those in the original tc-ds633.txt\n'
                  'File autogenerated using HGP633data_to_burnman.py\n'
                  '"""\n\n'
                  'from ..mineral import Mineral\n'
                  'from ..processchemistry import dictionarize_formula, formula_mass\n\n')

    outfile.write('"""\n'
                  'ENDMEMBERS\n'
                  '"""\n\n')

    def pprint_ordered_dict(d, leading_string, extra_whitespace=0):
        whitespace = ' ' * (len(leading_string)+2+extra_whitespace)
        s = pprint.pformat(d)
        s = s.replace('), ', ',\n{0}\''.format(whitespace))
        s = s.replace('\', ', '\': ').replace(' \'(\'', '\'')
        s = s.replace('OrderedDict([(', leading_string+'{').replace(')])', '}')
        return s

    formula = '0'
    for i in range(int(ds[0][0])):
        mbr = ds[i * 4 + 3][0]
        M = getmbr(ds, mbr)
        if mbr == 'and':  # change silly abbreviation
            mbr = 'andalusite'

        # Print parameters
        if hasattr(M, 'params'):
            outfile.write('class {0} (Mineral):\n'.format(mbr)+
                          '    def __init__(self):\n')

            s = pprint_ordered_dict(M.params, leading_string = '        self.params = ')
            s = s.replace('000000.0', 'e6')
            outfile.write(s)
            outfile.write('\n')

            # Print property modifiers (if they exist)
            if hasattr(M, 'landau_hp'):
                outfile.write('        self.property_modifiers = [[')
                s = pprint_ordered_dict(M.landau_hp, leading_string = '\'landau_hp\', ', extra_whitespace = 36)
                outfile.write(s)
                outfile.write(']]\n')

            if hasattr(M, 'bragg_williams') and M.bragg_williams['factor'] > 0:
                outfile.write('        self.property_modifiers = [[')
                s = pprint_ordered_dict(M.bragg_williams, leading_string = '\'bragg_williams\', ', extra_whitespace = 36)
                outfile.write(s)
                outfile.write(']]\n')


            outfile.write('        Mineral.__init__(self)\n\n')



    outfile.write('def cov():\n'
                  '    \"\"\"\n'
                  '    A function which loads and returns the variance-covariance matrix of the\n'
                  '    zero-point energies of all the endmembers in the dataset.\n\n'
                  '    Returns\n'
                  '    -------\n'
                  '    cov : dictionary\n'
                  '        Dictionary keys are:\n'
                  '        - endmember_names: a list of endmember names, and\n'
                  '        - covariance_matrix: a 2D variance-covariance array for the\n'
                  '          endmember zero-point energies of formation\n'
                  '    \"\"\"\n\n'
                  '    from .HGP_2018_ds633_cov import cov\n'
                  '    return cov\n\n')

# Process uncertainties
with open('HGP_2018_ds633_cov.py', 'wb') as outfile:

    outfile.write('# This file is part of BurnMan - a thermoelastic and '
                  'thermodynamic toolkit for the Earth and Planetary Sciences\n'
                  '# Copyright (C) 2012 - 2019 by the BurnMan team, '
                  'released under the GNU \n# GPL v2 or later.\n\n\n'
                  '"""\n'
                  'HGP_2018 (ds-633) zero-point energy covariance matrix\n'
                  'Derived from Holland, Green and Powell (2018) and references therein\n'
                  'Dataset version 6.33\n'
                  'The values in this document are all in S.I. units,\n'
                  'unlike those in the original tc-ds633.txt\n'
                  'File autogenerated using HGP633data_to_burnman.py\n'
                  '"""\n\n'
                  'from numpy import array\n\n'
                  'cov = ')


    import numpy as np
    n_mbrs = int(ds[0][0])

    names = []
    for i in range(n_mbrs):
        names.append(ds[i*4+3][0])

    cov = []
    for i in range(n_mbrs*4+4, len(ds)-2):
        cov.extend(map(float, ds[i]))

    i_utr = np.triu_indices(n_mbrs)
    i_ltr = np.tril_indices(n_mbrs)
    M = np.zeros((n_mbrs, n_mbrs))

    M[i_utr] = cov[1:]
    M[i_ltr] = M.T[i_ltr]

    M = M*1.e6 # (kJ/mol)^2 -> (J/mol)^2

    d = {'endmember_names':names,
         'covariance_matrix':M}

    np.set_printoptions(threshold='nan')

    import pprint
    pp = pprint.PrettyPrinter(indent=0, width=160, depth=3, stream=outfile)
    pp.pprint(d)

    outfile.write('\n')



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


from __future__ import absolute_import
import numpy as np
from .material import Material
from .averaging_schemes import AveragingScheme


class Model(object):

    """
    planetary model realization

    initialize with a rock, averaging scheme and list of pressures and temperatures

    then access any of the members to get lists of properties for pressures and temperatures

    all computations are done automatically and lazily

    """

    def __init__(self, rock, p, T, avgscheme):
        assert(len(p) == len(T))
        assert(len(p) > 0)
        assert(isinstance(rock, Material))
        assert(isinstance(avgscheme, AveragingScheme))

        self.rock = rock
        self.p = p
        self.T = T
        self.avgscheme = avgscheme

        self.moduli = None
        self.mat_rho = None
        self.mat_vs = None
        self.mat_vp = None
        self.mat_vphi = None
        self.mat_K = None
        self.mat_G = None
        self.mat_vp = None
        self.mat_vs = None
        self.mat_vphi = None
        self.alpha = None
        self.c_p = None
        self.c_v = None
        self.mat_V = None

    def v_s(self):
        self.compute_velocities_()
        return self.mat_vs

    def v_p(self):
        self.compute_velocities_()
        return self.mat_vp

    def v_phi(self):
        self.compute_velocities_()
        return self.mat_vphi

    def density(self):
        self.avg_moduli_()
        return self.mat_rho

    def K(self):
        self.avg_moduli_()
        return self.mat_K

    def G(self):
        self.avg_moduli_()
        return self.mat_G

    def thermal_expansivity(self):
        if self.alpha is None:
            self.calc_moduli_()
            n_pressures = len(self.p)
            self.alpha = np.zeros(n_pressures)
            for idx in range(n_pressures):
                V = np.array([m['V'] for m in self.moduli[idx]])
                alpha = np.array([m['alpha'] for m in self.moduli[idx]])
                self.alpha[
                    idx] = self.avgscheme.average_thermal_expansivity(V, alpha)

        return self.alpha

    def molar_heat_capacity_p(self):
        self.calc_heat_capacities_()
        return self.c_p

    def molar_heat_capacity_v(self):
        self.calc_heat_capacities_()
        return self.c_v

    def calc_moduli_(self):
        """
        Internal function to compute the moduli if necessary.
        """
        if self.moduli is None:
            self.moduli = [[] for p in self.p]

            for idx in range(len(self.p)):
                self.rock.set_state(self.p[idx], self.T[idx])
                (minerals, fractions) = self.rock.unroll()
                for (mineral, fraction) in zip(minerals, fractions):
                    e = {}
                    e['fraction'] = fraction
                    e['V'] = fraction * mineral.molar_volume
                    e['K'] = mineral.adiabatic_bulk_modulus
                    e['G'] = mineral.shear_modulus
                    e['rho'] = mineral.molar_mass / mineral.molar_volume
                    e['alpha'] = mineral.thermal_expansivity
                    e['c_v'] = mineral.molar_heat_capacity_v
                    e['c_p'] = mineral.molar_heat_capacity_p
                    self.moduli[idx].append(e)

    def avg_moduli_(self):
        """
        Internal function to average moduli if necessary.
        """
        if self.mat_V is None:
            self.calc_moduli_()
            n_pressures = len(self.p)
            self.mat_V = np.zeros(len(self.p))
            self.mat_K = np.zeros(len(self.p))
            self.mat_G = np.zeros(len(self.p))
            self.mat_rho = np.zeros(len(self.p))

            for idx in range(n_pressures):
                mods = self.moduli[idx]

                fractions = np.array([e['fraction'] for e in mods])
                V_frac = np.array([m['V'] for m in mods])
                K_ph = np.array([m['K'] for m in mods])
                G_ph = np.array([m['G'] for m in mods])
                rho_ph = np.array([m['rho'] for m in mods])

                self.mat_V[idx] = sum(V_frac)
                self.mat_K[idx] = self.avgscheme.average_bulk_moduli(
                    V_frac, K_ph, G_ph)
                self.mat_G[idx] = self.avgscheme.average_shear_moduli(
                    V_frac, K_ph, G_ph)
                self.mat_rho[idx] = self.avgscheme.average_density(
                    V_frac, rho_ph)

    def calc_heat_capacities_(self):
        """
        Internal function to compute the heat capacities if necessary.
        """
        if self.c_p is None:
            self.calc_moduli_()
            self.c_v = np.zeros(len(self.p))
            self.c_p = np.zeros(len(self.p))
            for idx in range(len(self.p)):
                fractions = np.array([m['fraction'] for m in self.moduli[idx]])
                alphas = np.array([m['alpha'] for m in self.moduli[idx]])
                c_v = np.array([m['c_v'] for m in self.moduli[idx]])
                c_p = np.array([m['c_p'] for m in self.moduli[idx]])
                self.c_v[idx] = self.avgscheme.average_heat_capacity_v(
                    fractions, c_v)
                self.c_p[idx] = self.avgscheme.average_heat_capacity_p(
                    fractions, c_p)

    def compute_velocities_(self):
        """
        Internal function to compute the velocities if necessary.
        """
        if self.mat_vp is None:
            self.avg_moduli_()
            self.mat_vs = np.ndarray(len(self.p))
            self.mat_vp = np.ndarray(len(self.p))
            self.mat_vphi = np.ndarray(len(self.p))

            for i in range(len(self.p)):
                self.mat_vs[i] = np.sqrt(self.mat_G[i] / self.mat_rho[i])
                self.mat_vp[i] = np.sqrt(
                    (self.mat_K[i] + 4. / 3. * self.mat_G[i]) / self.mat_rho[i])
                self.mat_vphi[i] = np.sqrt(self.mat_K[i] / self.mat_rho[i])



# BurnMan - a lower mantle toolkit
# Copyright (C) 2012, 2013, Heister, T., Unterborn, C., Rose, I. and Cottaar, S.
# Released under GPL v2 or later.

"""
RS_2014_liquids
^^^^^^^^^^^^^^^

Liquids from Ramo and Stixrude (2014) FPMD simulations.
There are some typos in the article which have been corrected where marked
with the help of David Munoz Ramo.
"""

from burnman import constants
from ..mineral import Mineral
from ..solidsolution import SolidSolution
from ..solutionmodel import *

from ..processchemistry import dictionarize_formula, formula_mass

class Fe2SiO4_liquid(Mineral):
    def __init__(self):
        formula = 'Fe2SiO4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Fe2SiO4_liquid',
            'formula': formula,
            'equation_of_state': 'dks_l',
            'V_0':  59.7717e-6, # modified for T_0
            'T_0':  1900., # corrected
            'O_theta': 1 ,
            'O_f': 4 ,
            'm': 0.6 ,
            'a': np.array([[-4252948.0, 997810.188],
                           [-599315.125, 12032.8936],
                           [12572739., 7299239.5],
                           [53442800.0, -26791676.0],
                           [52981912.0, 0.]]), # corrected order
            'zeta_0': 0.0161350928 , # 0.0166734, # the comment is a refit to David's dataset
            'xi': 0.34431043 , # 0.34431053, # the comment is a refit to David's dataset
            'Tel_0': 1919.3553 , # 1921.6813, # the comment is a refit to David's dataset
            'eta': 0.0127067110 , # 0.0127067, # the comment is a refit to David's dataset
            'spin_a': [-0.00011134, 0.00010863],
            'spin_b': [3.53793, -3.81421, 2.83703, -0.676241],
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}
        Mineral.__init__(self)
